/* @(#)rdmonit.c	14.1.1.1 (ES0-DMD) 09/16/99 09:22:23 */
/*===========================================================================
  Copyright (C) 1995 European Southern Observatory (ESO)
 
  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public 
  License along with this program; if not, write to the Free 
  Software Foundation, Inc., 675 Massachusetss Ave, Cambridge, 
  MA 02139, USA.
 
  Corresponding concerning ESO-MIDAS should be addressed as follows:
	Internet e-mail: midas@eso.org
	Postal address: European Southern Observatory
			Data Management Division 
			Karl-Schwarzschild-Strasse 2
			D 85748 Garching bei Muenchen 
			GERMANY
===========================================================================*/

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT   (c) 1994 European Southern Observatory
.IDENTIFIER  RDMONIT
.LANGUAGE    C
.AUTHOR      K. Banse, IPG-ESO Garching
.KEYWORDS    DeAnza unit
.PURPOSE     get contents of DeAnza monitor into a file
.ALGORITHM   straight forward
.INPUT/OUTPUT:
    call as  RDMON_C( dsplay, ichan, ochan, isplit, datdaz )

  input:
         int dsplay  :		device id
         int ichan   :		channel no. for image channel
         int ochan   :		channel no. for overlay channel
         int isplit  :		split screen flag

  output:
         int *datdaz :		array for screen pixels

.RETURNS     nothing
.COMMENTS    none
.ENVIRONment MIDAS
             #include <midas_def.h>   Prototypes for MIDAS interfaces
             #include <idinumd.h>     Global variables for DISPLAY interfaces

.VERSIONS    1.00       940517  F2C, taken from RDMONIT.FOR   RvHees
------------------------------------------------------------*/

/* Define _POSIX_SOURCE to indicate that this is a POSIX program */

#define  _POSIX_SOURCE 1

#include <midas_def.h>
#include <idinumd.h>

/*

*/

/*++++++++++++++++++++++++++++++
.IDENTIFIER  UNPACK
.PURPOSE     
.INPUT/OUTPUT:
    call as  UNPACK( cdata, size, zoom, ibuf )
  input:
         unsigned char *cdata:	pixel array in bytes
         int   size   :		dimanesion of cdata
         int   zoom   :		zoom factor

  output:
         int   *idata :		integer image data
         
.RETURNS     nothing
.COMMENTS    static function
--------------------------------*/
#ifdef __STDC__
      static void UNPACK( unsigned char *cdata, int size, int zoom, int *idata)
#else
      static void UNPACK( cdata, size, zoom, idata )
      unsigned char *cdata;
      int size, zoom, *idata;
#endif

{
register int n, nn;

int *work;
      
if ( zoom <= 1 )     
   I1UNPA_C( cdata, size, idata );
else
   {
   work = (int *) osmmget( size * sizeof( int ));
   I1UNPA_C( cdata, size, work );

   for (n=0; n<size ; n++)
      {
      for (nn=0; nn<zoom; nn++) *idata++ = work[n];
      }
   (void) osmmfree( (char *) work );
   }
}

/*

*/

/*++++++++++++++++++++++++++++++
.IDENTIFIER  RDCHN
.PURPOSE     reads a 2-dimensional image from single display memory 
.INPUT/OUTPUT:
    call as  RDCHN( dsplay, chan, itty, igr, imap, bgn, izm, data )
  input:
         int   dsplay :		Image Display unit no.
         int   chan   :         channel no. for image channel
         int   itty   :         ITT flag (1 or 0)
         int   igr    :         overlay channel 1:yes 0:no
         int   imap[4]:         channel subimage to be read
         int   bgn[2] :         start corner to write in DATA
         int   izm[2] :         replication of subimage

  output:
         int   *data  :         output buffer
         
.RETURNS     nothing
.COMMENTS    static function
--------------------------------*/
#ifdef __STDC__
      static void RDCHN( int dsplay, int chan, int itty, int igr, int *imap,
                                               int *bgn, int *izm, int *data )
#else
      static void RDCHN( dsplay, chan, itty, igr, imap, bgn, izm, data )
      int   dsplay, chan, itty, igr, *imap, *bgn, *izm, *data; 
#endif

{
register int ii, izy, lc, ll, nn;

unsigned char *cdata;
int  npix, nx, ny, ysta, *idata;


ll = (bgn[1] - 1) * QDSZX + bgn[0];		/* calculate offset */


/* read data from memory */

nx = imap[1] - imap[0] + 1;
ny = imap[3] - imap[2] + 1;
npix = nx * izm[0];
cdata = (unsigned char *) osmmget( nx * sizeof( unsigned char ));
idata = (int *) osmmget( npix * sizeof( int ));


/* read data  window */

ysta = imap[2];                 /* real y-offset within image display memory */
for (nn=0; nn<ny; nn++,ysta++)
   {
   (void) IIMSTW_C( dsplay, chan, LOADDR, nx, 1, QMDEP, *imap, ysta );
   (void) IIMRMY_C( dsplay, chan, nx, *imap, ysta, QMDEP, 4, itty, cdata );


   UNPACK(cdata,nx,izm[0],idata);	/* unpack + "zoom" into I*4 buffer */


   /* replicate vertically */   

   if (igr == 0)                                          /* image data */
      { 
      for (izy=0; izy<izm[1]; izy++)
         {
         lc = ll;
         for (ii=0; ii<npix; ii++) data[lc++] = cdata[ii];
         ll += QDSZX;
         }
      }

   else                                                   /* graphic data */
      {
      for ( izy = 0; izy < izm[1]; izy++ )
         {
         lc = ll;
         for ( ii = 0; ii < npix; ii++ )
            {
            if ( cdata[ii] != 0 ) data[lc] = cdata[ii];
            lc++;
            }
         ll += QDSZX;
         }
      }
   }

(void) osmmfree( cdata );
(void) osmmfree( (unsigned char *) idata );

}

/*

*/

void RDMON_C( dsplay, ichan, ochan, isplit, datdaz )
int dsplay, ichan, ochan, isplit, *datdaz;

{
int actvals, knul, num, unit;

int ioff[2], izm[2], imap[4];
int scrtop, ittflg, chanl;
int memszx, memszy;
int ix, iy, ixn[4], iyn[4], ixk[4], iyk[4];
int splcx[4][5], splcy[4][5], chans2[2], chansz[2];
int ige;
      
static int chiuse[4][5] = { { 1, 2, 3, 4, 2 },
                            { 1, 0, 0, 3, 4 },
                            { 0, 0, 1, 4, 0 },
                            { 0, 2, 3, 0, 0 }
                          };


memszx = QMSZX - 1;
memszy = QMSZY - 1;
scrtop = QDSZY - 1;


/* branch according to split mode */

if (isplit == 0)                                          /* NO split mode */
   {
   (void) DCGICH( ichan );

   ix = QDSZX / ZOOMX;
   iy = QDSZY / ZOOMY;

   *imap = SCROLX;                                 /* get x pixel at (0,0) */
   while ( *imap > NSX ) *imap -= memszx;

   if ( *imap < 0 )     
      {
      *ioff = -ZOOMX * (*imap) + 1;
      *imap = 0;
      }
   else
      *ioff = 1;

/* also check that we do not read too much for output array */

   if ( (ix * ZOOMX + *ioff - 1) > QDSZX ) ix = (QDSZX - *ioff + 1) / ZOOMX;
   imap[1] = *imap + ix - 1;

   imap[2] = SCROLY - scrtop / ZOOMY;              /* get y pixel at (0,0) */
   while ( imap[2] > NSY ) imap[2] -= memszy;

   if ( imap[2] < 0 )     
      {
      ioff[1] = -ZOOMY * imap[2] + 1;
      imap[2] = 0;
      }
   else
      ioff[1] = 1;

/* also check that we do not read too much for output array */

   if ( (iy * ZOOMY + ioff[1] - 1) > QDSZY) 
      iy = (QDSZY - ioff[2] + 1) / ZOOMY;
   imap[3] = imap[2] + iy - 1;

   izm[0] = ZOOMX;
   izm[1] = ZOOMY;

   ittflg = ITTYES;
   RDCHN( dsplay, ichan, ittflg, 0, imap, ioff, izm, datdaz );
   }

else                                                       /* YES split mode */
   {
   SPLCNT_C( splcx, splcy );
      
   if ( isplit <= 1 )     
      {
      chansz[0] = QDSZX / 2;
      chansz[1] = QDSZY / 2;
      }
   else if ( isplit <= 3 )
      {
      chansz[0] = QDSZX / 2;
      chansz[1] = QDSZY;
      }
   else 
      {
      chansz[0] = QDSZX;
      chansz[1] = QDSZY / 2;
      }
   chans2[0] = chansz[0] / 2;
   chans2[1] = chansz[1] / 2;

   /* read displayed subimage of ImageDisplay channels */

   for (num=0; num<4; num++)
      {
      chanl = chiuse[num][isplit] - 1;           /* chanl = channel no. */
      if ( chanl == -1 ) break;

      (void) DCGICH( chanl );			/* pull in channel info */

      ix = chansz[0] / ZOOMX;
      iy = chansz[1] / ZOOMY;
      ixn[chanl] = splcx[chanl][isplit] - chansz[0] / 2;
      ixk[chanl] = ixn[chanl] + chansz[0] - 1;

      imap[0] = SCROLX + ixn[chanl] / ZOOMX;             /* get x pixel */
      while ( *imap > NSX ) *imap -= memszx;

      if ( *imap < 0 )     
         {
         *ioff = ixn[chanl] - *imap * ZOOMX + 1;
         *imap = 0;
         }
      else
         *ioff = ixn[chanl] + 1;

      /* also check that we do not read too much for output array */

      if ( ZOOMX * ix + *ioff - 1 > ixk[chanl] + 1 ) 
         ix = (ixk[chanl] - *ioff + 2) / ZOOMX;
      imap[1] = *imap + ix - 1;

      iyn[chanl] = splcy[chanl][isplit] - chansz[1] / 2;
      iyk[chanl] = iyn[chanl] + chansz[1] - 1;

      imap[2] = SCROLY - (scrtop - iyn[chanl]) / ZOOMY;  /* get y pixel */
      while ( imap[2] > NSY ) imap[2] -= memszy;

      if ( imap[2] < 0 )     
         { ioff[1] = iyn[chanl] - imap[2] * ZOOMY + 1;
         imap[2] = 0;
         }
      else
         ioff[1] = iyn[chanl] + 1;

      /* also check that we do not read too much for output array */

      if ( ZOOMY * iy + ioff[1] - 1 > iyk[chanl] + 1 ) 
              iy = (iyk[chanl] - ioff[1] + 2) / ZOOMY;
      imap[3] = imap[2] + iy - 1;

      izm[0] = ZOOMX;
      izm[1] = ZOOMY;

      ittflg = ITTYES;
      RDCHN( dsplay, chanl, ittflg, 0, imap, ioff, izm, datdaz );
      }
   }


/* common overlay section - check if graphics enabled */

(void) SCKRDI( "DAZHOLD", 9, 1, &actvals, &ige, &unit, &knul );
if (ige == 1)
   {
   (void) DCGICH( ochan );

   if ( ZOOMX == 0 ) ZOOMX = 1;
   if ( ZOOMY == 0 ) ZOOMY = 1;

   ix = QDSZX / ZOOMX;
   iy = QDSZY / ZOOMY;

   *imap = SCROLX;                                 /* get x pixel at (0,0) */
   while ( *imap > NSX ) *imap -= memszx;

   if ( *imap < 0 )     
      {
      *ioff = -ZOOMX * (*imap) + 1;
      *imap = 0;
      }
   else
      *ioff = 1;

   /* also check that we do not read too much for output array */

   if ( ix * ZOOMX + *ioff - 1 > QDSZX ) ix = (QDSZX - *ioff + 1) / ZOOMX;
   imap[1] = *imap + ix - 1;

   imap[2] = SCROLY - scrtop / ZOOMY;              /* get y pixel at (0,0) */
   while ( imap[2] > NSY ) imap[2] -= memszy;

   if ( imap[2] < 0 )
      {
      ioff[1] = -ZOOMY * imap[2] + 1;
      imap[2] = 0;
      }
   else
      ioff[1] = 1;

   /* also check that we do not read too much for output array */

   if ((iy*ZOOMY + ioff[1] - 1) > QDSZY)
      iy = (QDSZY- ioff[1] + 1) / ZOOMY;
   imap[3] = imap[2] + iy - 1;

   izm[0] = ZOOMX;
   izm[1] = ZOOMY;
   RDCHN( dsplay, ochan, 0, 1, imap, ioff, izm, datdaz );
   }
}

